library angular2.src.core.compiler.dynamic_component_loader;

import "package:angular2/di.dart"
    show Key, Injector, Injectable, ResolvedBinding;
import "compiler.dart" show Compiler;
import "directive_metadata_reader.dart" show DirectiveMetadataReader;
import "package:angular2/src/facade/lang.dart"
    show Type, BaseException, stringify, isPresent;
import "package:angular2/src/facade/async.dart" show Future;
import "package:angular2/src/core/annotations/annotations.dart" show Component;
import "package:angular2/src/core/compiler/view_factory.dart" show ViewFactory;
import "package:angular2/src/core/compiler/view_hydrator.dart"
    show AppViewHydrator;
import "element_injector.dart" show ElementRef, DirectiveBinding;
import "view.dart" show AppView; /**
 * @exportedAs angular2/view
 */

class ComponentRef {
  ElementRef location;
  dynamic instance;
  AppView componentView;
  Function _dispose;
  ComponentRef(ElementRef location, dynamic instance, AppView componentView,
      Function dispose) {
    this.location = location;
    this.instance = instance;
    this.componentView = componentView;
    this._dispose = dispose;
  }
  get injector {
    return this.location.injector;
  }
  get hostView {
    return this.location.hostView;
  }
  dispose() {
    this._dispose();
  }
} /**
 * Service for dynamically loading a Component into an arbitrary position in the internal Angular
 * application tree.
 *
 * @exportedAs angular2/view
 */
@Injectable()
class DynamicComponentLoader {
  Compiler _compiler;
  ViewFactory _viewFactory;
  AppViewHydrator _viewHydrator;
  DirectiveMetadataReader _directiveMetadataReader;
  DynamicComponentLoader(Compiler compiler,
      DirectiveMetadataReader directiveMetadataReader, ViewFactory viewFactory,
      AppViewHydrator viewHydrator) {
    this._compiler = compiler;
    this._directiveMetadataReader = directiveMetadataReader;
    this._viewFactory = viewFactory;
    this._viewHydrator = viewHydrator;
  } /**
   * Loads a component into the location given by the provided ElementRef. The loaded component
   * receives injection as if it in the place of the provided ElementRef.
   */
  Future<ComponentRef> loadIntoExistingLocation(Type type, ElementRef location,
      [Injector injector = null]) {
    this._assertTypeIsComponent(type);
    var annotation = this._directiveMetadataReader.read(type).annotation;
    var componentBinding = DirectiveBinding.createFromType(type, annotation);
    return this._compiler.compile(type).then((componentProtoView) {
      var componentView = this._viewFactory.getView(componentProtoView);
      this._viewHydrator.hydrateDynamicComponentView(
          location, componentView, componentBinding, injector);
      var dispose = () {
        throw new BaseException("Not implemented");
      };
      return new ComponentRef(location,
          location.elementInjector.getDynamicallyLoadedComponent(),
          componentView, dispose);
    });
  } /**
   * Loads a component in the element specified by elementOrSelector. The loaded component receives
   * injection normally as a hosted view.
   */
  Future<ComponentRef> loadIntoNewLocation(
      Type type, ElementRef parentComponentLocation, dynamic elementOrSelector,
      [Injector injector = null]) {
    this._assertTypeIsComponent(type);
    return this._compiler.compileInHost(type).then((hostProtoView) {
      var hostView = this._viewFactory.getView(hostProtoView);
      this._viewHydrator.hydrateInPlaceHostView(
          parentComponentLocation, elementOrSelector, hostView, injector);
      var newLocation = hostView.elementInjectors[0].getElementRef();
      var component = hostView.elementInjectors[0].getComponent();
      var dispose = () {
        this._viewHydrator.dehydrateInPlaceHostView(
            parentComponentLocation, hostView);
        this._viewFactory.returnView(hostView);
      };
      return new ComponentRef(
          newLocation, component, hostView.componentChildViews[0], dispose);
    });
  } /**
   * Loads a component next to the provided ElementRef. The loaded component receives
   * injection normally as a hosted view.
   */
  Future<ComponentRef> loadNextToExistingLocation(
      Type type, ElementRef location, [Injector injector = null]) {
    this._assertTypeIsComponent(type);
    return this._compiler.compileInHost(type).then((hostProtoView) {
      var hostView = location.viewContainer.create(-1, hostProtoView, injector);
      var newLocation = hostView.elementInjectors[0].getElementRef();
      var component = hostView.elementInjectors[0].getComponent();
      var dispose = () {
        var index = location.viewContainer.indexOf(hostView);
        location.viewContainer.remove(index);
      };
      return new ComponentRef(
          newLocation, component, hostView.componentChildViews[0], dispose);
    });
  } /** Asserts that the type being dynamically instantiated is a Component. */
  _assertTypeIsComponent(Type type) {
    var annotation = this._directiveMetadataReader.read(type).annotation;
    if (!(annotation is Component)) {
      throw new BaseException(
          '''Could not load \'${ stringify ( type )}\' because it is not a component.''');
    }
  }
}
